from typing import Optional,  Literal

from tzafon import Computer

from .utils import singleton, get_logger
from .constants import Settings

logger = get_logger(__name__)
config = Settings()

@singleton
class TzafonClient:
    """Singleton class to ensure only one Tzafon client instance exists."""
    def __init__(self, api_key: Optional[str] = None):
        """
        Initialize the TzafonClient.

        Args:
           api_key: The API key for Tzafon. If not provided, it will be retrieved from the environment variable `TZAFON_API_KEY`.
        """
        self._api_key = api_key or config.api_key.get_secret_value()
        self.computer = None

       
    def initialize(self, kind: Literal["browser", "desktop"] = "browser") -> Computer:
        """
        Initialize the Tzafon client and create a computer instance.

        Args:
            kind: The type of computer to create. Can be "browser" or "desktop". Defaults to "browser".

        Returns:
            The initialized Tzafon computer instance.

        Raises:
             ValueError: If the API key is missing.
        """
        if not self._api_key:
            logger.error("Tzafon API key is required. Read here on how to get your api key: https://docs.tzafon.ai/quickstart#get-your-api-key")
            raise ValueError("Tzafon API key is required. Read here on how to get your api key: https://docs.tzafon.ai/quickstart#get-your-api-key")


        client = Computer(api_key=self._api_key)

        self.computer = client.create(kind=kind)

        logger.info("Tzafon client initialized")
        logger.info("Synchronous computer created with id %s", self.computer.id)
    
        return self.computer

    def terminate(self):
        """Terminate the Tzafon computer instance."""
        self.computer.terminate()
    
    


import logging
import sys

def get_logger(name: str) -> logging.Logger:
    """
    Get a configured logger for the given module name.
    
    Args:
        name: The name of the module (typically __name__)
        
    Returns:
        A configured logging.Logger instance
    """
    logger = logging.getLogger(name)
    
    if not logger.handlers:
        handler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(logging.INFO)
        
    return logger

def singleton(cls):
    """
    A decorator that implements the Singleton design pattern.
    
    Ensures that a class has only one instance and provides a global point of 
    access to it. Subsequent calls to the decorated class will return the 
    same instance.
    
    Args:
        cls: The class to be decorated.
        
    Returns:
        The singleton instance of the class.
    """
    instances = {}
    def get_instance(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]
    return get_instance


from typing import (
    Iterator, 
    AsyncIterator,
    Optional, 
    Sequence, 
    Union
    )

from langchain_core.documents import Document
from langchain_core.document_loaders.base import BaseLoader
from playwright.sync_api import sync_playwright
from playwright.async_api import async_playwright

from langchain_tzafon.core import TzafonClient
from langchain_tzafon.utils import get_logger
from langchain_tzafon.constants import Settings

logger = get_logger(__name__)

config = Settings()

class TzafonLoader(BaseLoader):
    """
    Load pre-rendered web pages using a headless browser hosted on Tzafon.

    This loader uses the Tzafon service to render web pages and extract their content.
    It supports both synchronous and asynchronous loading.

    Attributes:
        urls (List[str]): List of URLs to load.
        api_key (Optional[str]): The Tzafon API key.
        text_content (bool): Whether to extract only text content (True) or full HTML (False).
    """
    def __init__(
        self,
        urls: Union[str, Sequence[str]],
        api_key: Optional[str] = None,   
        text_content: Optional[bool] = True,  
    ):
        """
        Initialize the TzafonLoader.

        Args:
            urls: A single URL or a list of URLs to load.
            api_key: The Tzafon API key. If not provided, it will be retrieved from the `TZAFON_API_KEY` environment variable.
            text_content: If True, extracts text content. If False, extracts the raw HTML. Defaults to True.
        """

        if isinstance(urls, str):
            self.urls = [urls]
        else:
            self.urls = urls

        self.api_key = api_key or config.api_key.get_secret_value()
        self.text_content = text_content

        self.client = TzafonClient(api_key=self.api_key)


    def lazy_load(self) -> Iterator[Document]:
        """
        Load pages from URLs synchronously.

        Yields:
             Iterator[Document]: An iterator of Document objects containing the page content.
        """
        computer = self.client.initialize()
        cdp_url = f"{config.api_base_url}/computers/{computer.id}/cdp?token={self.api_key}"

        try:
            with sync_playwright() as playwright:
                browser = playwright.chromium.connect_over_cdp(cdp_url)
                context =  browser.contexts[0] if browser.contexts else browser.new_context()
                for url in self.urls:
                    page = context.new_page()
                    
                    page.goto(url)
                    if self.text_content:
                        page_text = page.inner_text("body")
                        content = str(page_text)
                    else:
                        page_html = page.content()
                        content = str(page_html)

                    page.close()
                browser.close()

                yield Document(
                    page_content=content,
                    metadata={
                        "url": url,
                    },
                )
        except Exception as e:
            logger.error(f"Error loading page: {e}")
            raise
        finally:
            computer.terminate()

    async def alazy_load(self) -> AsyncIterator[Document]:
        """
        Load pages from URLs asynchronously.

        Yields:
             AsyncIterator[Document]: An async iterator of Document objects containing the page content.
        """
        computer = self.client.initialize()
        cdp_url = f"{config.api_base_url}/computers/{computer.id}/cdp?token={self.api_key}"

        try:
            async with async_playwright() as playwright:
                browser = await playwright.chromium.connect_over_cdp(cdp_url)
                context = browser.contexts[0] if browser.contexts else await browser.new_context()
                
                for url in self.urls:
                    page = await context.new_page()

                    await page.goto(url)
                    if self.text_content:
                        page_text = await page.inner_text("body")
                        content = str(page_text)
                    else:
                        page_html = await page.content()
                        content = str(page_html)

                    await page.close()
                await browser.close()

                yield Document(
                    page_content=content,
                    metadata={
                        "url": url,
                        },
                    )
        except Exception as e:
            logger.error(f"Error loading page: {e}")
            raise
        finally:
            computer.terminate()


import os

from pydantic_settings import BaseSettings
from pydantic import SecretStr

from .utils import singleton

@singleton
class Settings(BaseSettings):
    """
    Pydantic settings for the Tzafon LangChain integration.
    
    Attributes:
        api_base_url (str): The base URL for the Tzafon API. Defaults to "https://api.tzafon.ai".
    """
    api_base_url: str = "https://api.tzafon.ai"
    api_key: SecretStr =  SecretStr(os.getenv("TZAFON_API_KEY", default=""))
   



import pytest
from unittest.mock import MagicMock, patch, AsyncMock
from langchain_tzafon.tzafon_loader import TzafonLoader

@pytest.fixture
def mock_settings():
    with patch("langchain_tzafon.tzafon_loader.config") as mock_config:
        mock_config.api_base_url = "http://mock-api"
        mock_config.api_key.get_secret_value.return_value = "default_key"
        yield mock_config

@pytest.fixture
def mock_client():
    with patch("langchain_tzafon.tzafon_loader.TzafonClient") as MockClient:
        mock_instance = MockClient.return_value
        mock_browser = MagicMock()
        mock_browser.id = "computer-123"
        mock_instance.initialize.return_value = mock_browser
        yield MockClient

@pytest.fixture
def mock_sync_playwright():
    with patch("langchain_tzafon.tzafon_loader.sync_playwright") as mock_pw:
        yield mock_pw

@pytest.fixture
def mock_async_playwright():
    with patch("langchain_tzafon.tzafon_loader.async_playwright") as mock_pw:
        yield mock_pw

def test_initialization(mock_settings, mock_client):
    # Test 1: Single URL
    loader = TzafonLoader(urls="http://example.com", api_key="test_key")
    assert loader.urls == ["http://example.com"]
    assert loader.api_key == "test_key"
    mock_client.assert_called_with(api_key="test_key")
    mock_client.return_value.initialize.assert_called()
    assert loader.browser == mock_client.return_value.initialize.return_value

    # Test 2: List of URLs
    loader = TzafonLoader(urls=["http://a.com", "http://b.com"], api_key="test_key")
    assert loader.urls == ["http://a.com", "http://b.com"]

def test_lazy_load_text(mock_settings, mock_client, mock_sync_playwright):
    # Setup Playwright mocks
    mock_pw_context = mock_sync_playwright.return_value.__enter__.return_value
    mock_browser = mock_pw_context.chromium.connect_over_cdp.return_value
    mock_context = mock_browser.new_context.return_value
    mock_page = mock_context.new_page.return_value
    
    # Setup Page behavior
    mock_page.inner_text.return_value = "Mock Page Content"
    
    # Execute
    loader = TzafonLoader(urls="http://example.com", api_key="test_key")
    docs = list(loader.lazy_load())
    
    # Assertions
    assert len(docs) == 1
    assert docs[0].page_content == "Mock Page Content"
    assert docs[0].metadata["url"] == "http://example.com"
    
    # Verify Playwright calls
    mock_pw_context.chromium.connect_over_cdp.assert_called_with(
        "http://mock-api/computers/computer-123/cdp?token=test_key"
    )
    mock_page.goto.assert_called_with("http://example.com")
    mock_page.inner_text.assert_called_with("body")
    mock_page.close.assert_called()
    mock_browser.close.assert_called()
    mock_client.return_value.initialize.return_value.terminate.assert_called()

def test_lazy_load_html(mock_settings, mock_client, mock_sync_playwright):
    mock_pw_context = mock_sync_playwright.return_value.__enter__.return_value
    mock_page = mock_pw_context.chromium.connect_over_cdp.return_value.new_context.return_value.new_page.return_value
    mock_page.content.return_value = "<html><body>Mock HTML</body></html>"
    
    loader = TzafonLoader(urls="http://example.com", api_key="test_key", text_content=False)
    docs = list(loader.lazy_load())
    
    assert len(docs) == 1
    assert docs[0].page_content == "<html><body>Mock HTML</body></html>"
    mock_page.content.assert_called()

@pytest.mark.asyncio
async def test_alazy_load_text(mock_settings, mock_client, mock_async_playwright):
    # Setup Async Playwright Mock
    mock_pw_context = mock_async_playwright.return_value.__aenter__.return_value
    
    mock_browser = AsyncMock()
    mock_pw_context.chromium.connect_over_cdp.return_value = mock_browser
    
    mock_context = AsyncMock()
    mock_browser.new_context.return_value = mock_context
    
    mock_page = AsyncMock()
    mock_context.new_page.return_value = mock_page
    
    mock_page.inner_text.return_value = "Async Mock Page Content"
    
    # Execute
    loader = TzafonLoader(urls="http://example.com", api_key="test_key")
    
    docs = []
    async for doc in loader.alazy_load():
        docs.append(doc)
        
    # Assertions
    assert len(docs) == 1
    assert docs[0].page_content == "Async Mock Page Content"
    
    # Verify Async Calls
    mock_pw_context.chromium.connect_over_cdp.assert_awaited_with(
            "http://mock-api/computers/computer-123/cdp?token=test_key"
    )
    mock_page.goto.assert_awaited_with("http://example.com")
    mock_page.inner_text.assert_awaited_with("body")
    mock_page.close.assert_awaited()
    mock_browser.close.assert_awaited()
    mock_client.return_value.initialize.return_value.terminate.assert_called()
